#!/usr/bin/perl
# pixlog.pl 5 June 2000 Matt Post

# This script sits on a FIFO that syslog opens and pumps firewall
# logging data to. This fifo must exist before the syslog daemon starts.
#
# To cause it to write the statistics file, create $PRINTFILE. If
# the file contains any nonzero number in it, this program will 
# reset all internal variables. Otherwise, it will just write out
# what it has and continue to accumulate.
#
# Give it any arguments, and nothing will be printed to STDOUT
# (I use 'quiet')

##########################
# CONFIGURABLE VARIABLES #
##########################

# Adjust this to the location of 
my $fifo = "/var/log/pix/info";

# This file, when present, causes the statistics to be written to the
# html file.
my $PRINTFILE = "/var/log/pix/.print";

# base html directory, where all the statistics files should be placed
my $HTML_DIR = "/usr/local/share/apache/htdocs/pix";

#################################
# END OF CONFIGURABLE VARIABLES #
#################################

use FileHandle;				# makes file IO easier
use POSIX qw( strftime );		# time stuff
use CGI qw( -no_debug :standard );	# for html output
use Net::hostent;			# ip -> hostname conversion
use strict;				# make programming harder
use vars qw( %telnet %o_traffic %o_traffic_rec %deny %deny_rec
	$faddr $faddr_port $laddr $laddr_port 
	$linecount $parsecount $start $end );

# if any arguments are given, nothing will be printed to STDOUT
my $arg = shift;

# don't print entries with fewer than this number of entries
my $THRESHOLD = 5;

my $OUTFILE = "$HTML_DIR/" . (strftime "%m%d%Y", localtime) . "-info.html";
my $SYMLINK = "$HTML_DIR/00LATEST.html";


# Make sure the file exists as a fifo
unless (-p $fifo)  {
	unlink $fifo
		or die "$fifo is not a pipe $!";
	# use 'and' because system calls return backwards
	system('mknod', $fifo, 'p')
		and die "can't mknod $fifo $!";
}

print "Process ID: $$\n"
	unless $arg;

#############
# main loop # open the socket, and keep on reading/parsing
#############

# initialize everything, including recording the start time
reset_vars();
 
while (1)  {
	print "** Opening $fifo\n"
		unless $arg;
	my $ifh = new FileHandle $fifo, "r"
		or die "open $fifo $!";
	while (<$ifh>)  {
		# number of lines coming through
		$linecount++;

		next if (/UDP/);		# who really cares?

		# number of lines coming through that we parse
		$parsecount++;

		# check for the existence of $PRINTFILE, which tells us to
		# write out the statisticts. Any nonzero content in this
		# file indicates that everything should be reset afterwards
		# whether to start over
		if (-e $PRINTFILE)  {
			my $pfh = new FileHandle $PRINTFILE, "r"
				or die "can't open $PRINTFILE $!";
			# everything's reset if there's anything nonzero
			chomp( my $val = <$pfh>);
			write_stats($val);
		}

		SWITCH:  {

		   # TCP denied
		   /PIX-6-106015/ and do  {
			@_ = (split);
			($faddr,$faddr_port) = (split /\//, $_[14]);
			($laddr,$laddr_port) = (split /\//, $_[16]);

			print "Denied: $faddr/$faddr_port -> $laddr/$laddr_port\n"
				unless $arg;
			$deny{to}{$laddr}{$laddr_port}++;
			$deny_rec{hosts}{$laddr}++;
			$deny_rec{ports}{$laddr_port}++;

			last SWITCH;
		   };

		   # Built TCP connection
		   /PIX-6-302001/ and do  {
			# foreign/local addresses
			($faddr, $faddr_port) = (split /\//, (split)[-5]);
			($laddr, $laddr_port) = (split /\//, (split)[-1]);

			# skip inbound connections
			next unless /Built outbound/;

			print "Outbound: $laddr/$laddr_port -> $faddr/$faddr_port\n"
				unless $arg;
			$o_traffic{to}{$faddr}{$faddr_port}++;
			$o_traffic_rec{hosts}{$faddr}++;
			$o_traffic_rec{ports}{$faddr_port}++;

			last SWITCH;
		   };

		   # Permitted telnet session
		   /PIX-6-307002/ and do  {
			my ($host) = (split)[-1];	# last field
			print "Telnet success: $host\n"
					unless $arg;
			$telnet{good}{$host}++;
			last SWITCH;
		   };

		   # Telnet login session bad
		   /PIX-6-307003/ and do  {
			my ($host,$count) = (split)[14..15];
			$count =~ s/\(//;
			print "Telnet fail: $host ($count)\n"
				unless $arg;
			$telnet{bad}{$host} += $count;
			last SWITCH;
		   };

		   # Bad 'enable' password
		   /PIX-6-308001/ and do  {
			my ($host,$count) = (split)[-1,-6];
			$host =~ s/\)+//;

			print "Enable fail: $host ($count)\n"
				unless $arg;
			$telnet{enable_fail}{$host} += $count;
			last SWITCH;
		  };
		}
	}
	print "** Closing $fifo\n"
		unless $arg;
	undef $ifh;
}


###########################
#      SUBROUTINES        #
###########################

sub write_stats   {
	my $date = strftime "%A, %e %B %Y", localtime;
	my $key;
	my $cgi = new CGI;
	my $startover = shift;

	# ACK! Make sure that you calculate the time right when you write
	# the file, and not a long time before that
	my $OUTFILE = "/usr/local/share/apache/htdocs/pix/" . (strftime "%m%d%Y", localtime) . "-info.html";
	
	# record the end of the time range for this logging
	$end = localtime();

	print "** Opening $OUTFILE\n"
		unless $arg;
	my $ofh = new FileHandle $OUTFILE, "w"
		or die "Can't open $OUTFILE $!";
	my $old_fh = select $ofh;
	$ofh->autoflush(1);

	print $cgi->start_html(-title => "Resnet firewall stats for $date",
		-bgcolor => "#ffffff");

	print "<h1>Resnet firewall stats for $date</h1>\n\n";


	############################
	# Firewall Telnet Sessions #
	############################
	# figure out whether to print header (only print if there are 
	# sub-categories)
	if ($telnet{good} or $telnet{bad} or $telnet{enable_fail})  {
		print "<hr width=200 align=left>\n";
		print "<h3>Firewall Telnet Sessions</h3>\n";

		print <<EOF;
<table cellspacing=2 cellpadding=4>
<tr>
	<th>From host</th>
	<th>Count</th>
	<th></th>
</tr>
EOF
	}

	  # print this one?
	if ($telnet{good})  {
	  my $firsttime = 1;
	  foreach $key (reverse sort 
	    { $telnet{good}{$a} <=> $telnet{good}{$b} }
	    keys %{ $telnet{good} })  {
		print "<tr>\n\t<td bgcolor='#eeeeee'>$key</td>\n";
		print "\t<td bgcolor='#66cc33'>$telnet{good}{$key}</td>\n";
		if ($firsttime)  {
			print "\t<td><i>Successful logins</i></td>\n";
			$firsttime = 0;
		}
		print "</tr>\n";
	  }
	}

	# only print if there are entries
	if ($telnet{bad})  {
	  my $firsttime = 1;
	  foreach $key (reverse sort { $telnet{bad}{$a} <=> $telnet{bad}{$b} }
	   keys %{ $telnet{bad} })  {
		print "<tr>\n\t<td bgcolor='#eeeeee'>$key</td>\n";
		print "\t<td bgcolor='#ff3300'>$telnet{bad}{$key}</td>\n";
		if ($firsttime)  {
			print "\t<td><i>Failed logins</i></td>\n";
			$firsttime = 0;
		}
		print "</tr>\n";
	  }
	}

	# only print sub-category if there are entries
	if ($telnet{enable_fail})  {
	  my $firsttime = 1;
	  foreach $key (reverse sort 
	     { $telnet{enable_fail}{$a} <=> $telnet{enable_fails}{$b} }
	     keys %{ $telnet{enable_fail} })  {
		print "<tr>\n\t<td bgcolor='#eeeeee'>$key</td>\n";
		print "\t<td bgcolor='#ff3300'>$telnet{enable_fail}{$key}</td>\n";
		if ($firsttime)  {
			print "\t<td><i>Failed enable attempts</i></td>\n";
			$firsttime = 0;
		}
		print "</tr>\n";
	  }
	}

	if ($telnet{good} or $telnet{bad} or $telnet{enable_fail})  {
	  print "</table>\n\n";
	}

	####################
	# Outbound Traffic #
	####################
	# only print header if things are defined
	if ($o_traffic_rec{hosts} or $o_traffic_rec{ports})  {
 	  print "<hr width=200 align=left>\n";
	  print "<h3>Outbound Traffic</h3>\n";
	}

	# print hosts?
	if ($o_traffic_rec{hosts})  {
	  print "\n<table cellspacing=2 cellpadding=4>\n";
	  print <<EOF;
<tr>
	<th align="left">Hostname</th>
	<th>IP Address</th>
	<th>Total packet count</th>
	<th>Port</th>
	<th>Packet count</th>
</tr>
EOF
	  # first, sort (descending) by host
	  # %o_traffic_rec contains a per-port access count. We use this to
	  # to print the highest port counts at the top, for each port
	  # The per-host port counts are contained in %o_traffic
	  foreach $key (reverse sort 
	     { $o_traffic_rec{hosts}{$a} <=> $o_traffic_rec{hosts}{$b} }
	     keys %{ $o_traffic{to} })  {
		# don't print records under the threshold
		next if ($o_traffic_rec{hosts}{$key} < $THRESHOLD);

		print "<tr>";
		if (my $h = gethost($key))  {
			my $name = $h->name;
			print "\t<td bgcolor='#eeeeee'>$name</td>\n\t<td bgcolor='#eeeeee'>$key</td>\n\t<td bgcolor='#e1e1ae'>$o_traffic_rec{hosts}{$key}</td>\n";
		} else  {
			print "\t<td bgcolor='#eeeeee'>???</td>\n\t<td bgcolor='#eeeeee'>$key</td>\n\t<td bgcolor='#e1e1ae'>$o_traffic_rec{hosts}{$key}</td>\n";
		}
			
		# then, sort by ports within that host
		my $firsttime = 1;

		# $printed is used to keep track of whether any port info
		# was printed. If nothing was (say, all information was under
		# the threshold), we need to close the table information to
		# prevent gaps and ugly html. Used again right after this loop.
		my $printed = 0;

		foreach my $key2  (reverse sort 
		   { $o_traffic{to}{$key}{$a} <=> $o_traffic{to}{$key}{$b} }
		   keys %{ $o_traffic{to}{$key} })  {
			# don't print records under the threshold
			next if ($o_traffic{to}{$key}{$key2} < $THRESHOLD);

			# mark that something has been printed
			$printed = 1;

			unless ($firsttime)  {
				print "<tr>\t<td colspan=3 bgcolor='#eeeeee'><br></td>\n";
			}
			$firsttime = 0;

			print "\t<th bgcolor='#c3c390'>$key2</th>\n"; 
			print "\t<td bgcolor='#c3c390'>$o_traffic{to}{$key}{$key2}</td>\n";

			print "</tr>\n";
		}
		# make sure that some port information was printed (if not,
		# we need to close the <tr> tag and print an informative 
		# message).
		unless ($printed)  {
			print "\t<td colspan=2 align='center'><i>Nothing above threshold</i></td>\n</tr>\n";
		}
	  }
	  print "<tr>\n\t<td colspan=5></td>\n</tr>\n";
	}
	print "</table>\n\n";

	# only print if there are some entries
	if ($o_traffic_rec{ports})  {
	  print "<hr width=200 align=left>\n";
	  print "<h3>Outbound Ports</h3>\n";
	  print "<table cellspacing=2 cellpadding=4>\n";
	  print "<tr>\n";
	  for (1..4)  {
		print "\t<th>Port</th>\n\t<th>Packet count</th>\n";
	  }
	  print "</tr>\n";

	  # print in 4 columns, using $col for formatting
	  my $col = 1;
	  foreach $key (reverse sort 
	    { $o_traffic_rec{ports}{$a} <=> $o_traffic_rec{ports}{$b} }
	    keys %{ $o_traffic_rec{ports} })  {
		# don't print records under the threshold
		next if ($o_traffic_rec{ports}{$key} < $THRESHOLD);

		print "<tr>\n" if ($col == 1);

		print "\t<th align=left bgcolor='#eeeeee'>$key</th>\n";
		print "\t<td align=left bgcolor='#c3c390'>$o_traffic_rec{ports}{$key}</td>\n";

		print "</tr>\n" if ($col == 4);

		$col = $col % 4 + 1;
	  }
	  print "</table>\n\n\n";
	}


	##############
	# TCP Denied #
	##############
	# only print category if sub-categories have entries
	if ($deny_rec{hosts} or $deny_rec{ports})  {
	  print "<hr width=200 align=left>\n";
	  print "<h3>Inbound Traffic (denied)</h3>\n";
	}

	# only print if there are entries
	if ($deny_rec{hosts})  {
	  print "<table cellspacing=2 cellpadding=4>\n";
	  print <<EOF;
<tr>
	<th align="left">IP Address</th>
	<th>Total packet count</th>
	<th>Port</th>
	<th>Packet count</th>
</tr>
EOF
	  # first, sort (descending) by host
	  # %deny_rec contains a per-port access count. We use this to
	  # print the highest port counts at the top, for each host. The
	  # per-host port counts are contained in %deny
	  foreach $key (reverse sort 
	     { $deny_rec{hosts}{$a} <=> $deny_rec{hosts}{$b} } 
	     keys %{ $deny{to} }) {
		# don't print records under the threshold
		next if ($deny_rec{hosts}{$key} < $THRESHOLD);

		print "<tr>\n\t<td bgcolor='#eeeeee'>$key</th>\n";
		print "\t<td bgcolor='#e1e1ae'>$deny_rec{hosts}{$key}</td>\n";

		# then, sort by ports within that host

		# $printed is used to keep track of whether any port info
		# was printed. If nothing was (say, all information was under
		# the threshold), we need to close the table information to
		# prevent gaps and ugly html. Used again right after this loop.
		my $printed = 0;
		my $firsttime = 1;
		foreach my $key2 (reverse sort
		   { $deny{to}{$key}{$a} <=> $deny{to}{$key}{$b} }
		   keys %{ $deny{to}{$key} })  {
			next if ($deny{to}{$key}{$key2} < $THRESHOLD);

			# mark that something has been printed
			$printed = 1;

			unless ($firsttime)  {
				print "<tr>\n\t<td colspan=2 bgcolor='#eeeeee'><br></td>\n";
			}
			$firsttime = 0;

			print "\t<th bgcolor='#c3c390'>$key2</th>\n";
			print "\t<td bgcolor='#c3c390'>$deny{to}{$key}{$key2}</td>\n";

			print "</tr>\n";
		}

		# make sure that some port information was printed (if not,
		# we need to close the <tr> tag and print an informative 
		# message).
		unless ($printed)  {
			print "\t<td colspan=2 align='center'><i>Nothing above threshold</i></td>\n</tr>\n";
		}
	  }

	  print "<tr>\n\t<td colspan=4></td>\n</tr>\n";
	  print "</table>\n\n";
	}

	if ($deny_rec{ports})  {
	  print "<hr width=200 align=left>\n";
	  print "<h3>Inbound Ports</h3>\n";
	  print "<table cellspacing=2 cellpadding=4>\n";
	  print "<tr>\n";
	  for (1..4)  {
		print "\t<th>Port</th>\n\t<th>Packet count</th>\n";
	  }
	  print "</tr>\n";

	  # print in four columns, using $col for formatting
	  my $col = 1;
	  foreach $key (reverse sort 
	    { $deny_rec{ports}{$a} <=> $deny_rec{ports}{$b} }
	    keys %{ $deny_rec{ports} }) {
		# don't print records under the threshold
		next if ($deny_rec{ports}{$key} < $THRESHOLD);

                print "<tr>\n" if ($col == 1);

		print "\t<th align=left bgcolor='#eeeeee'>$key</th>\n";
		print "\t<td align=left bgcolor='#c3c390'>$deny_rec{ports}{$key}</td>\n";

                print "</tr>\n" if ($col == 4);

		$col = $col % 4 + 1;
	  }
	  print "</table>\n";
	}

	print <<EOF;
<br><br>
<font size=2>Script <i>$0</i> started $start and completed $end, receiving $linecount lines of data and parsing $parsecount of them.</font>
EOF

	print $cgi->end_html;

	# close filehandles
	undef $cgi;
	undef $ofh;
	select $old_fh;

	# reset everything if $PRINTFILE contained nonzero data
	if ($startover)  {
		reset_vars();
	}

	# delete the file that told us to print
	unlink $PRINTFILE;

	# make the symlink so we can just go to 00LATEST.html
	unlink $SYMLINK if (-e $SYMLINK);
	symlink $OUTFILE, $SYMLINK;

	return;
}


# Resets all the statistics variables. Basically, everything from
# the 'use vars qw//' line at top
sub reset_vars  {
	print "** Resetting vars\n"
		unless $arg;
	%deny = %deny_rec = %telnet = %o_traffic = %o_traffic_rec = ();
	$faddr = $faddr_port = $laddr = $laddr_port 
	  = $linecount = $parsecount = 0;

	# record the starting time
	$start = localtime();
}
